/* Final Fill 'Er Up Operation Code.
   This code controls and encompasses all programming of all 3 boards and their respective interfaces for the entire proper machine use
   David Finnell, Eric Hawkins, Richard McManus, Michael Slusarczyk*/

#include <Arduino.h>
#include <SPI.h>
#include <Wire.h>
#include <Adafruit_GFX.h>
#include <Adafruit_SSD1306.h>
#include <VL6180X.h>
#include <TFT_eSPI.h>
#include "SparkFun_TCA9534.h"
#include "TLC59108.h"
#include "Final_Frontier_28.h"
#include "NDEE.h"
#include "FillErUpLogo.h"

// Function Declarations
int singleHeightMeasure();
bool CheckCupPresent(void);
float liquidSensorHeightMeasure(float cupHeightMeasured, bool first_occassion);
bool writeReg(uint8_t reg, const void *pBuf, size_t size);
uint8_t readReg(uint8_t reg, const void *pBuf, size_t size);

// Keurig Functions
void stopBrew();
void restartBrew();
bool pumpRunning();
bool monitorBrewButton();
void brewSteadyState();
bool CheckLimitSwitch ();

//Fun LED Functions
void SexyLights();
void SexyLightsShort();

// Rotary Encoder Functions
float DialUIInteraction(int cupHeight);
void DialUISetup();
bool ButtonCheck(void);
float AverageLiquidHeight(float);

// Multiplexor Functions
void mux_GPIO_HIGH(int mux_num, int GPIO_Pin);
void mux_GPIO_LOW(int mux_num, int GPIO_Pin);

//Display Functions
void DisplayFillCup (void);
void DisplayBrewInProcess (void);
void DisplayBrewComplete (void);
void DisplayBrewCompleteEnjoy(void);
void DisplayNeedMoreWater(void);
void DisplayCupRemoval(void);
void DisplayActiveFillRate(int dots, float liquidHeight, float targetHeight, float bottomHeight);
void DisplayHeatingInProgress(int dots);
void DisplayAddPod(void);
void DisplayLogo(void);
void DisplayNDEE(void);
void WaitStartUp(void);
void WaitRestart(void);
void DisplayPowerOn(void);
void DisplayBrewAborted(void);
void DisplayMeasuring(void);

// Rotary Encoder IO Pins
#define ENC_CLK 16   
#define ENC_DT 15
#define ENC_SW 18

// LED Shows Liquid Filled
#define LiquidLED 4

// Cup Height LEDs
#define LIMPOS 4
#define MOS 5
#define BREW 2
#define RES 6
#define POW 1
#define PUMP 17

// I2C Communication
#define SDA 14
#define SCL 13
#define SDA1 40
#define SCL1 39
#define NUM2AVG 3
#define PUMPRUNTIME 14000

// Code for multiplexor
TCA9534 tof_Mux[2]; // Create an array of two muxes for the two boards
bool mux_GPIO_State[2][8]= { {LOW, LOW, LOW, LOW, LOW, LOW, LOW, LOW}, {LOW, LOW, LOW, LOW, LOW, LOW, LOW, LOW} };
float liquidHeights[NUM2AVG];
int liquidHeightsCounter = 0;


// ToF Sensor Array
VL6180X sensors_vl6180; // Treat 16x as a single device

// SEN0590 Definitions
#define address 0x74
#define MaxHeight = 27.5
uint8_t buf[2] = {0};
uint8_t dat = 0xB0;

// Button Definitions
unsigned long lastDebounceTime0 = 0;
unsigned long debounceDelay0 = 750;
unsigned long lastDebounceTime1 = 0;
unsigned long debounceDelay1 = 750;



// *********** SETUP *************


float cupHeightMeasured;
int LEDHeightCounter;
int numLiquid = 0;
float heightOutput = 0.0;
bool noButton = true;
bool firstPress = true;
float percent;

// Circular Display
TFT_eSPI tft = TFT_eSPI(); // Invokes custom library for our GC9A01 TFT
int counter;
int currentStateCLK;
int lastStateCLK;
unsigned long lastButtonPress = 0;

// Led Driver
#define ADDR_LED0 0x40
#define ADDR_LED1 0x41

TLC59108 leds0(ADDR_LED0);  //Invokes class for the two separate LED drivers from Library for the LED
TLC59108 leds1(ADDR_LED1);  // Drivers on each daughter board

int mux_pins = 8;
int num_mux = 2;

int mux_num;
int gpio;

int restart_counter;

void setup()
{
  restart_counter = 0;
  Serial.begin(115200);
  delay(1000);

  Serial.println("Entered Setup"); 

  Wire.begin(SDA, SCL);   //I2C Bus Initialization for Daughter Board Comms
  Wire1.begin(SDA1, SCL1); // I2C Bus Initialization for Liquid Height Sensor

  // Setup LED Drivers
  leds0.init();
  leds1.init();
  leds0.setLedOutputMode(TLC59108::LED_MODE::PWM_IND);
  leds1.setLedOutputMode(TLC59108::LED_MODE::PWM_IND);
 
 
  // Set up Multiplexors and their Pins
  for(mux_num=0; mux_num < num_mux; mux_num++){
    while(tof_Mux[mux_num].begin(Wire, 0x20 + mux_num) == false){
      Serial.println("Mux not connected");
      delay(500);
      };
    for(gpio = 0; gpio < mux_pins; gpio++){
      tof_Mux[mux_num].pinMode(gpio, GPIO_OUT); //Configure all pins to output
      tof_Mux[mux_num].digitalWrite(gpio, LOW); //Set all GPIOs to LOW
    }
  }

  // Set up Time-of-Flight Sensors
  Serial.println("Start of ToF Config loop");

  // Configuration of both muxes and initialization of all TOF sensors
  for(mux_num = num_mux-1; mux_num > -1; mux_num--){
    for (gpio = mux_pins-1; gpio > -1; gpio--){

      // Print Current ToF and Mux Being Detected
      Serial.print("GPIO: ");
      Serial.println(gpio);
      Serial.print("    Mux: ");
      Serial.println(mux_num);


      // Enable one TOF, configure, then disable
      mux_GPIO_HIGH(mux_num, gpio);

      sensors_vl6180.init();
      sensors_vl6180.setAddress(0x29);
      sensors_vl6180.setTimeout(100);  // Set timeout to 500ms

      mux_GPIO_LOW(mux_num, gpio);

    }
  }
 
  SexyLights();
  tft.init(); //Initializes our UI Screen

  Serial.println("Waiting for Height Measurement");

  //Everything is initialized and we can now enter loop
}



// ************* LOOP  ************** //

float currLiquid = 0;
float firstLiquid = 0;
int maxRunTime = 45000;
int numTargetHeight = 0;
bool first = true;
bool notFirst = false;
bool aborted;

void loop(){
  for (int element = 0; element < NUM2AVG; element++){ //Initializes averaging algorithm for liquid height sensing to start at 0
    liquidHeights[element] = 0;
  }
  bool filled = false; //Cup is not filled
  aborted = false; //Brew process is not aborted
  bool cup_present = true; //Assuming a cup is present
  bool firstTime = true;
  percent = 0;
  DialUISetup(); 
  DisplayPowerOn();
  while (pumpRunning() == true){} //Checks the pump voltage as a way of seeing if coffee maker is plugged in
  //The UI will prompt the user to plug in the Keurig before proceeding until the power is sensed to be sent to the Keurig
  WaitStartUp(); //"Keurig Sleep Mode", panning through logos until a cup is seen under the coffee maker
  DisplayMeasuring(); //Displays that the cup is being measured
  SexyLightsShort(); //All LEDs blink twice to indicate cup is measuring
  LEDHeightCounter = singleHeightMeasure(); // Complete all ToF Measurements & Trigger LED & Calculate Cup Height via # of LEDS on
  brewSteadyState(); // Configure Keurig Pins to their proper IO for a successfuul brew process
  cupHeightMeasured = DialUIInteraction(LEDHeightCounter); //Allows the user to adjust LED fill height until they insert the K-CUP and press
  //the brew button. Sends back a float in cm for the Liquid Height Sensor to look for
  Serial.print("Height of Liquid to Be Poured: ");
  Serial.print(cupHeightMeasured); // Print Height of Liquid to be Poured
  Serial.println(" cm");
  counter = 0;
  // Enters while loop when brewing/heating process has begun
  while (!filled && !aborted && cup_present){ // Outer while loop only finishes when brew is finished, aborted, or cup is removed
    if (ButtonCheck() == true){ //Repeatedly checks to see if User wants to stop brew by pushing the dial button
      aborted = true; 
      break;
    }
    currLiquid = liquidSensorHeightMeasure(cupHeightMeasured,notFirst); // Repetedly checks height reported from liquid height sensor
    Serial.print("Current Level: ");
    Serial.println(currLiquid);
    cup_present = CheckCupPresent(); //Repeadtedly checks if a cup is still present
    if (counter == 4){ //Determines how many dots will be displayed on the Heating function
      counter = 0;
    }
    DisplayHeatingInProgress(counter);
    counter++;
   
    if (currLiquid > cupHeightMeasured){
      filled = true;
    } //Determines that if the liquid height is above the target liquid height, to terminate the brew 

    if (pumpRunning() == true) // Checks that pump is off before proceeding into next stage of brew
    {

      //active display percentage
      int timeStart = millis();
      int runTime = 0;

      firstLiquid = liquidSensorHeightMeasure(cupHeightMeasured, firstTime); //Does an initial cup height measurement so as to give a reference for the fill %

      while (runTime < maxRunTime && !filled) // Enters if runtime is under 45 seconds and measured liquid level is not above desired height
      {
        if (ButtonCheck() == true){ //New loop so repeatedly checks to make sure user didn't abort brew by pressing dial
          aborted = true;
          break;
        }
        if (CheckCupPresent() == false){ //Repeatedly Checks that Cup is still present
          cup_present = false;
          break;
        }
        counter++;
        if (counter == 4){ //Determines how many dots will be displayed on Pouring UI animaton
          counter = 0;
        }
        
        runTime = millis() - timeStart;
        currLiquid = liquidSensorHeightMeasure(cupHeightMeasured,notFirst); // Continuously Measure Liquid Height
        DisplayActiveFillRate(counter, currLiquid, cupHeightMeasured, firstLiquid); //Display the active fill percentage to user
        Serial.print("Runtime: ");
        Serial.println(runTime);
        Serial.print("Water Height: ");
        Serial.println(currLiquid);

       if (currLiquid > cupHeightMeasured){ //Tells ESP32 if cup is filled so it can terminate
        filled = true;
       }

        if(runTime>maxRunTime){ // If reservoir empty or runtime over and not filled, automatically starts new brew
          Serial.println("Restarting...");
          restartBrew();
          break;
        }
      }
    }
  }

  stopBrew(); //Once while loop is exited for whatever reason, powers off coffee maker 

  
  //Decides why the brew was stopped and does certain display given the case
  if (aborted)
  {
    DisplayBrewAborted();
    Serial.println("Brew Aborted....Check Water");
  }
  else if (!cup_present){
    Serial.println("Cup Removed...Brew Aborted");
    DisplayCupRemoval();
    while (ButtonCheck() == false){} //Waits until user resets the process by pressing the dial
  }
  else{
    Serial.println("Brew Completed... Reset Device to Complete Next Brew"); //Series of LED Light Displays and UI Displays to indicate successful brew!
    DisplayBrewComplete();
    SexyLights();
    DisplayBrewCompleteEnjoy();
    SexyLights();
    DisplayBrewComplete();
    SexyLights();
    WaitRestart();
  }
}


// *********** FUNCTION DEFINITIONS *************

//TOF Measurement Functions

int singleHeightMeasure()
/*Loops through all of the TOF sensors, reads from proper registers to make sure everything is operating properly. Then, the TOF's measure
the distances, if there is an object within a defines threshold of the sensors, it returns true and lights up the LED corresponding to the GPIO TOF that
saw the object there. It then finds the max TOF that went on to indicate an LED Height to the next function, and also turns off the top two TOFs
due to the overlap that occurs when measuring. Returns a proper cup measurement.
*/
{
  int distances[num_mux][8];
  int maxToFGPIO;
  int maxToFMux;
  int threshold = 30;
  // Read measurements for all sensors

    for(mux_num = num_mux-1; mux_num > -1; mux_num--){ // Do for both muxes

      for (gpio = mux_pins-1; gpio > -1; gpio--){

        Serial.print("GPIO: ");
        Serial.println(gpio);
        mux_GPIO_HIGH(mux_num, gpio); // Turn on GPIO
        Serial.println(sensors_vl6180.readReg(0x000),HEX);
        if (sensors_vl6180.timeoutOccurred()) { Serial.print(" TIMEOUT"); }
        sensors_vl6180.writeReg(0x018,0x01);
        sensors_vl6180.writeReg(0x02d,0x00);
        if (sensors_vl6180.timeoutOccurred()) { Serial.print(" TIMEOUT"); }
        delay(150);
        Serial.println(sensors_vl6180.readReg(0x04F),BIN);
        if (sensors_vl6180.timeoutOccurred()) { Serial.print(" TIMEOUT"); }

        // Complete Single ToF Measurement
        distances[mux_num][gpio] = sensors_vl6180.readReg(0x062);


        // Configure LEDs Measurements (Determine if threshold is met)
        if(distances[mux_num][gpio] < threshold && mux_num == 0){
          Serial.print("Blink");
          leds0.setBrightness(gpio,100);
          maxToFGPIO = gpio;
          maxToFMux = 0;
        }
        else if(distances[mux_num][gpio] < threshold && mux_num == 1){
          Serial.print("Blink");
          leds1.setBrightness(gpio,100);
          maxToFGPIO = gpio;
          maxToFMux = 1;

         
        }
        else if(distances[mux_num][gpio] > threshold && mux_num == 0){
          Serial.print("Off");
          leds0.setBrightness(gpio,0);
                 
        }
        else if(distances[mux_num][gpio] > threshold && mux_num == 1){
          Serial.print("Off");
          leds1.setBrightness(gpio,0);        
        }

        else{Serial.print("Error");}

        if (sensors_vl6180.timeoutOccurred()) { Serial.print(" TIMEOUT"); }

        Serial.print(" Distance: ");
        Serial.println(distances[mux_num][gpio]);

        if (sensors_vl6180.timeoutOccurred()) { Serial.print(" TIMEOUT"); }
       
        mux_GPIO_LOW(mux_num, gpio);
      }
    }

    Serial.println(maxToFGPIO);
    if (maxToFMux == 0){
      leds0.setBrightness(maxToFGPIO,0);
    }
    else if(maxToFMux == 1){
      leds1.setBrightness(maxToFGPIO,0);
    }
    // Determine Overall Cup Height
    int Sens_Meas_Height = -2; // Distance from bottom to first sensor
    for (int mux_num = 0; mux_num < 2; mux_num++){
      for (int gpio = 8; gpio > -1; gpio--)
      {
        if (distances[mux_num][gpio] < threshold){ // Cup Close Enough to Sensor
          Sens_Meas_Height ++; // Assume 1cm bottom + 1.27cm per sensor spacing
        }
      }
    }

  Serial.println(" ");
  Serial.println(" ");
  Serial.println(" ");
  Serial.print("Cup Height = ");
  Serial.println(Sens_Meas_Height);
  Serial.println(" ");
  Serial.println(" ");
  delay(250);
  return Sens_Meas_Height; // Return Height of Cup as an integer of LED height
}

bool CheckCupPresent(void){
  //Does a TOF measurement for only the bottom sensor in the same procedure as the above and returns whether or not
  //it saw something there
  int threshold = 30;
  int distance = 0;
  mux_num = 1;
  gpio = 7;
  bool presence = true;
  mux_GPIO_HIGH(mux_num, gpio); // Turn on GPIO
  Serial.println(sensors_vl6180.readReg(0x000),HEX);
  if (sensors_vl6180.timeoutOccurred()) { Serial.print(" TIMEOUT"); }
  sensors_vl6180.writeReg(0x018,0x01);
  sensors_vl6180.writeReg(0x02d,0x00);
  if (sensors_vl6180.timeoutOccurred()) { Serial.print(" TIMEOUT"); }
  delay(150);
  Serial.println(sensors_vl6180.readReg(0x04F),BIN);
  if (sensors_vl6180.timeoutOccurred()) { Serial.print(" TIMEOUT"); }

  distance = sensors_vl6180.readReg(0x062);
  if (sensors_vl6180.timeoutOccurred()) { Serial.print(" TIMEOUT"); }

  if (distance > threshold){
    presence = false;
    }
   mux_GPIO_LOW(mux_num, gpio);

   return presence;
}

//Liquid Height Sensing functions

float liquidSensorHeightMeasure(float cupHeightMeasured, bool first_occassion)
/*Initializes proper measurement offsets to complete a measurement from the Liquid Height Sensor and does proper address
  and buffer arithmetic to piece together the right number that the sensor is receiving, converts into centimeters and returns that value.
*/
{
  float curr_Liquid_Poured;
  float FillHeight = cupHeightMeasured; // Fill Height with 5% buffer from actual measured height
  float SEN_dist;

  float Sen_to_bottom = 29;

  // Initialize Summing & Iteration Variables
  float SEN_dist_raw = 0;
  int num_iter = 5;
  float high = 0;
  float distance;
  // Write to & Read from Liquid Sensor 4 Times
  for (int counter = 0; counter < num_iter+1; counter++)
  {
    writeReg(0x10, &dat, 1);
    delay(50);
    readReg(0x02, buf, 2);
    distance = buf[0] * 0x100 + buf[1] + 10;
    if (distance > high){high = distance;}
    SEN_dist_raw += distance; // Combine Last 4 Measurements using +=
    delay(5);
  }

  // Convert Raw Data to Height Measurement
  float SEN_dist_mm = (SEN_dist_raw-high) / float(num_iter); // Average Last "num_iter" measurements in mm
  SEN_dist = (SEN_dist_mm) / 10.0;             // Convert from mm to cm
  SEN_dist_raw = 0;                            // Reset averaging variable

 
  // Serial Plotter Output
  curr_Liquid_Poured = Sen_to_bottom - SEN_dist; // Total distance - sensor measurement = Liquid Height

  delay(25);

  if(first_occassion){
    return(curr_Liquid_Poured);
  }
  return AverageLiquidHeight(curr_Liquid_Poured);
}


// SEN0590 Functions: Liquid Height Sensors register read and write functions, drawn straight from product example
// for easy use of I2C buffers for getting the right value

uint8_t readReg(uint8_t reg, const void *pBuf, size_t size)
{
  if (pBuf == NULL)
  {
    Serial.println("pBuf ERROR!! : null pointer");
  }
  uint8_t *_pBuf = (uint8_t *)pBuf;
  Wire1.beginTransmission(address);
  Wire1.write(&reg, 1);
  if (Wire1.endTransmission() != 0)
  {
    return 0;
  }
  delay(20);
  Wire1.requestFrom((uint8_t)address, (uint8_t)size);
  for (uint16_t i = 0; i < size; i++)
  {
    _pBuf[i] = Wire1.read();
  }
  return size;
}

bool writeReg(uint8_t reg, const void *pBuf, size_t size)
{
  if (pBuf == NULL)
  {
    Serial.println("pBuf ERROR!! : null pointer");
  }
  uint8_t *_pBuf = (uint8_t *)pBuf;
  Wire1.beginTransmission(address);
  Wire1.write(&reg, 1);

  for (uint16_t i = 0; i < size; i++)
  {
    Wire1.write(_pBuf[i]);
  }
  if (Wire1.endTransmission() != 0)
  {
    return 0;
  }
  else
  {
    return 1;
  }
}


// Rotary Encoder Functions

float DialUIInteraction (int cupHeight) {
  /* Creates an LED offset from the SingleHeightMeasure Function based off of the cup size to reduce over-measurement from sensor beam-width overlap. Then lights up the
     proper LEDS given the offset created. Continuously lights up the number of LEDs given how much the user turns up or turns down the dial, giving them a visual of
     what they're adjusting the cup height to. Has Rotary Encoder code in there for checking the logic of the two signals to see if the position has changed. Does not leave loop
     until Limit Switch is opened and closd (K-Cup is inserted) along with the Brew Button being pressed after the limit switch. Lastly converts to centimeters based
     off the board dimensions to return to the chip what our target fill height is. 
  */
  // Read the current state of CLK
  DisplayFillCup();
  int iter;
  if (cupHeight < 7){
    counter = cupHeight - 2;
  }
  else{
    counter = cupHeight - 4;
  }
  if (counter<=8){
    for (iter = 0; iter <= 7; iter++){
      leds0.setBrightness(iter, 0);
    }
    for (iter = 0; iter < (8-counter); iter++){
      leds1.setBrightness(iter, 0);
    }
    for (iter = 7; iter >= (8-counter); iter--){
      leds1.setBrightness(iter, 100);
    }
  }
  else {
    for (iter = 7; iter>= 0; iter--){
      leds1.setBrightness(iter, 100);
    }
    for (iter = 7; iter >= (16-counter); iter--){
      leds0.setBrightness(iter, 100);
    }
    for (iter = 0; iter < (16-counter); iter++){
      leds0.setBrightness(iter, 0);
    }
  }
  counter--;
  int limCounter = 0;
  bool limSwitchCycle = false;
  bool brewButtonPressed = false;
  while (brewButtonPressed == false || limSwitchCycle == false){
    if (ButtonCheck() == true){
      aborted = true;
      break;
    }
    if(limCounter == 0 && CheckTopClosed() == true){
      limCounter = 1;
    }
    if(limCounter == 1 && CheckTopClosed() == false){
      limCounter = 2;
    }
    if(limCounter == 2){
      if(CheckTopClosed() == true){
        limSwitchCycle = true;
      }
      else{
        limSwitchCycle = false;
      }
    }
    if(limSwitchCycle == true){
      if(monitorBrewButton() == true){
        brewButtonPressed = true;
      }
    }
    //Serial.print("Counter:  ");
    //Serial.println(counter);
    currentStateCLK = digitalRead(ENC_CLK);
    // If last and current state of CLK are different, then pulse occurred
    // React to only 1 state change to avoid double count
    if (currentStateCLK != lastStateCLK  && currentStateCLK == 1){
      
      // If the DT state is different than the CLK state then
      // the encoder is rotating CCW so decrement
      if (digitalRead(ENC_DT) != currentStateCLK) {
        if (counter > 0){
          counter -= 1;
        }
      } else {
        if (counter < cupHeight-1){
          counter += 1;
        }
      }
      Serial.print("Counter:  ");
      Serial.println(counter); //Adjusts LEDs lit up based off User adjustment
      if (counter<=8){
        for (iter = 0; iter <= 7; iter++){
          leds0.setBrightness(iter, 0);
        }
        for (iter = 0; iter < (8-counter); iter++){
          leds1.setBrightness(iter, 0);
        }
        for (iter = 7; iter >= (8-counter); iter--){
          leds1.setBrightness(iter, 100);
        }
      }
      else {
        for (iter = 7; iter>= 0; iter--){
          leds1.setBrightness(iter, 100);
        }
        for (iter = 7; iter >= (16-counter); iter--){
          leds0.setBrightness(iter, 100);
        }
        for (iter = 0; iter < (16-counter); iter++){
          leds0.setBrightness(iter, 0);
        }
      }
    }
    
    // Remember last CLK state
    lastStateCLK = currentStateCLK;
    // Read the button state

      delayMicroseconds(100);
  }
  float fillHeight = 1.27*((float)counter)+4;
  Serial.print("Fill Height Set: ");
  Serial.println(fillHeight);
  return fillHeight;
}

void DialUISetup(void)
// Sets up the Rotary Encoder Pins and initializes the TFT Screen
{

  tft.init();

  // Set encoder pins as inputs
  pinMode(ENC_CLK, INPUT_PULLUP);
  pinMode(ENC_DT, INPUT_PULLUP);
  pinMode(ENC_SW, INPUT_PULLUP);

  // Setup Serial Monitor
  Serial.begin(115200);

  int lastStateCLK = digitalRead(ENC_CLK);
}

bool ButtonCheck(void){ //Simple check to see if the button has been pressed or not on the Rotary Encoder
  int btnState = digitalRead(ENC_SW);

  //If we detect LOW signal, button is pressed
  if (btnState == LOW) {
    //if 50ms have passed since last LOW pulse, it means that the
    //button has been pressed, released and pressed again
    if (millis() - lastButtonPress > 50) {
      return true;
      }

    // Remember last button press event
    lastButtonPress = millis();
  }

  // Put in a slight delay to help debounce the reading
  delay(1);
  return false;
}



// Keurig Functions

void brewSteadyState()

//Sets up the pins for a normal brew process to commence on our terms for Reading and Writing, 
//MOSFET going high corresponds to user control of the Limit Switch and Not us 
{
  // Define pinMode for Brew Control
  pinMode(LIMPOS, INPUT);
  pinMode(BREW, INPUT);
  pinMode(MOS, OUTPUT);
  pinMode(RES, OUTPUT);
  pinMode(POW, INPUT);
  pinMode(PUMP, INPUT);

  digitalWrite(MOS, HIGH);
  digitalWrite(RES, LOW);
  Serial.println("Brew Steady State.");
}

bool monitorBrewButton() 
/* Checks if Brew Button is pressed, if so follows the coffee maker in filling the heater to the proper amount based on timing, but also checks via perfectly mapped timing
  whether or not there is enough water to fill the entire heater, and prompts the brew to be aborted. If aborted due to lack of water, resets brew process for user and asks
  them to put in more water. Also continusouly checks during the initial runtime to ensure there is a cup present and the user didn't abort the process by pressing the dial. 
  Also sees if heater is full or not so the code knows whether or not the process starts with water pumping in or with the heater being on.
*/
{
  pinMode(BREW, INPUT);
  if (digitalRead(BREW) == 0)
  {
    Serial.println("Brew Button Pressed.");
    delay(500);
    if (pumpRunning()){
      DisplayBrewInProcess();
      Serial.println("Pump is running. Waiting 12s to open resevoir sensor.");
      int timeNow = millis();
      bool buttonPressed = false;
      while(millis() - timeNow < PUMPRUNTIME){
        if(ButtonCheck() == true || CheckCupPresent() == false){
          buttonPressed = true;
          aborted = true;
          break;
        }
      }
      pinMode(RES, INPUT);
      Serial.println("Resevoir Sensor Opened.");
      timeNow = millis();
      while(pumpRunning()==true){
        if(ButtonCheck() == true || CheckCupPresent() == false){
          buttonPressed = true;
          aborted = true;
          break;
        }
        Serial.println(millis()-timeNow);
        if(millis()-timeNow>4650){
          DisplayNeedMoreWater();
          Serial.println("Waiting for more water to be put in");
          pinMode(LIMPOS, INPUT);
          pinMode(BREW, INPUT);
          pinMode(MOS, OUTPUT);
          pinMode(RES, OUTPUT);
          pinMode(POW, INPUT);

          pinMode(RES, OUTPUT);
          delay(50);
        
          // Limit Switch
          Serial.println("Limit switch opened and closed.");
          pinMode(LIMPOS, OUTPUT);
          digitalWrite(LIMPOS, HIGH);
          delay(100);
          pinMode(LIMPOS, INPUT);
          while (monitorBrewButton() == false) {}
          break;
        }
      }
      delay(1000);
    }
    else{
      DisplayHeatingInProgress(0);
      Serial.println("Pump not running. Will proceed assuming heater is full.");
    }
    return true;
  }
  else
    return false;
}

bool pumpRunning()
//Checks if pump is on or off based off an analogRead because digital doesn't measure properly based on pull-ups/pull-downs on pump
{
  int sum = 0;
  for(int i = 0; i < 10; i++){
    sum+=analogRead(PUMP);
  }
  int val = sum/10;
  if (val < 3000)
  {
    return true;
  }
  return false;
}

void restartBrew()
/* Does the proper pin setups and processes for the Keurig State machine to automatically start another brew without any user intervention via our tap into 
   those components. If there is a lot of coffee left to brew, waits for user to put in another K-Cup. If pump is the first thing to run, it means that it's 
   trying to find more water, than therefore immediately stops and waits for the user to put in more water and re-activate the brew. 
   */
{
  restart_counter++;
  
  pinMode(LIMPOS, INPUT);
  pinMode(BREW, INPUT);
  pinMode(MOS, OUTPUT);
  pinMode(RES, OUTPUT);
  pinMode(POW, INPUT);

  // Power Button
  pinMode(POW, OUTPUT);
  digitalWrite(POW, LOW);
  delay(100);
  pinMode(POW, INPUT);
  Serial.println("Power button pressed. Brew aborted.");
  pinMode(RES, OUTPUT);
  delay(50);

  // Limit Switch
  Serial.println("Limit switch opened and closed.");
  pinMode(LIMPOS, OUTPUT);
  digitalWrite(LIMPOS, HIGH);
  delay(100);
  pinMode(LIMPOS, INPUT);
  delay(5000);

  if (cupHeightMeasured - currLiquid > 3.0){
    DisplayAddPod();
    while (monitorBrewButton() == false){}
    delay(500);
    restart_counter = 0;
  }
  else{
    pinMode(BREW, OUTPUT);
    delay(200);
    digitalWrite(BREW, LOW);
    delay(200);
    pinMode(BREW, INPUT);
    Serial.println("Brew button pressed. Brew Initiated");

    delay(500);
  }


  if (pumpRunning()){
    DisplayNeedMoreWater();
    Serial.println("Waiting for more water to be put in");
    while (monitorBrewButton() == false) {}
  }
  else{
    Serial.println("Pump not running. Will proceed assuming heater is full.");
  }
}

void stopBrew() //Taps into power button on coffee maker to stop the brew process
{
  // Power Button
  pinMode(POW, OUTPUT);
  digitalWrite(POW, LOW);
  delay(100);
  pinMode(POW, INPUT);
  Serial.println("Power button pressed. Brew aborted.");
  brewSteadyState();
  delay(50);
}

bool CheckTopClosed (void){
  //Checks if the Limit Switch has been opened/closed
  //Note that the logic is slightly changed for the limit switch since we had to do a last minute switch to a new coffee maker
  //which had opposite logic of what we had originally investigated since it was a newer board
  if (analogRead(LIMPOS)<1000){
    return true;
  }
  else{
    return false;
  }
}




// Mux Functions: The below 2 functions will set the GPIO of a Mux on a daughter board high or low. 

void mux_GPIO_HIGH(int mux_num, int GPIO_Pin){

  // Set MUX GPIO Output HIGH
  tof_Mux[mux_num].digitalWrite(GPIO_Pin, HIGH);
  mux_GPIO_State[mux_num][GPIO_Pin] = HIGH;

}

void mux_GPIO_LOW(int mux_num, int GPIO_Pin){
  // Set MUX GPIO Output LOW

  tof_Mux[mux_num].digitalWrite(GPIO_Pin, LOW);
  mux_GPIO_State[mux_num][GPIO_Pin] = LOW;

}

// Runs LEDS upward and downward repeatedly along with blinks all of them for a cool effect to provide in the system
void SexyLights(){
  Serial.println("Sexxxy Lights!!!");
  for(int numLasers = 0; numLasers < 3; numLasers++){
    for(int counter = 15; counter > -1; counter--){
      if(counter > 7){
        leds1.setBrightness(counter - 8, 100);
        delay(10);
        leds1.setBrightness(counter - 8, 0);
        delay(10);
      }
      else{
        leds0.setBrightness(counter, 100);
        delay(10);
        leds0.setBrightness(counter, 0);
        delay(10);
      }
     }
     
     for(int counter = 0; counter < 16; counter++){
      if(counter > 7){
        leds1.setBrightness(counter - 8, 100);
        delay(10);
        leds1.setBrightness(counter - 8, 0);
        delay(10);
      }
      else{
        leds0.setBrightness(counter, 100);
        delay(10);
        leds0.setBrightness(counter, 0);
        delay(10);
      }
     }
    }

   for(int numFlashes = 0; numFlashes < 3; numFlashes++){
     for(int counter = 0; counter < 16; counter++){
       if(counter > 7){
         leds1.setBrightness(counter - 8, 100);
       }
       else{
         leds0.setBrightness(counter, 100);
       }
     }
     delay(100);
     for(int counter = 0; counter < 16; counter++){
       if(counter > 7){
         leds1.setBrightness(counter - 8, 0);
       }
       else{
         leds0.setBrightness(counter, 0);
       }
     }
     delay(100);
   }
}

//FUN LED Functions
void SexyLightsShort(void){ //Shorter version of the above function with only two blinks.
    for(int numFlashes = 0; numFlashes < 3; numFlashes++){
    for(int counter = 0; counter < 16; counter++){
      if(counter > 7){
        leds1.setBrightness(counter - 8, 100);
      }
      else{
        leds0.setBrightness(counter, 100);
      }
    }
    delay(100);
    for(int counter = 0; counter < 16; counter++){
      if(counter > 7){
        leds1.setBrightness(counter - 8, 0);
      }
      else{
        leds0.setBrightness(counter, 0);
      }
    }
    delay(100);
  }
}

float AverageLiquidHeight(float liquidHeight){
//Function computes the moving average of the liquid height so as to filter out splashes toward the liquid sensor that could skew the
//measurement
  liquidHeights[liquidHeightsCounter] = liquidHeight;
  liquidHeightsCounter++;
  if(liquidHeightsCounter == NUM2AVG){
    liquidHeightsCounter = 0;
    }
  int i;
  float total;
  for (i = 0; i < NUM2AVG; i++){
    total += liquidHeights[i];
  }
  float averageLiquidHeight = total/NUM2AVG;
  return averageLiquidHeight;
}

//UI Display Functions

void DisplayFillCup (void){
  //UI Display Function to prompt the user to adjust the fill height according to the measurement and start the brew when ready
  tft.fillScreen(TFT_WHITE);
  tft.setTextSize(1);
  tft.loadFont(Final_Frontier_28);
  tft.setCursor(25, 30, 4);
  
  tft.setTextColor(TFT_RED);
  tft.println("    Cup Height");
  tft.println("         Measured");
  tft.setTextColor(TFT_BLUE);
  tft.println("       Turn dial to");
  tft.println("    adjust fill height");
  tft.setTextColor(TFT_DARKGREEN);
  tft.println("     Insert pod and");
  tft.println("     press K-button");
  tft.println("        when ready.");
  tft.unloadFont();
}

void DisplayBrewInProcess (void){ // UI Display Function initiates Yellow Screen Indicating the Brew Cycle has Started if the Brew Cycle starts pumping in water
  tft.setTextSize(1);
  tft.fillScreen(TFT_YELLOW);  
  tft.loadFont(Final_Frontier_28);
  tft.setCursor(50, 110, 4);
  
  tft.setTextColor(TFT_BLACK);
  tft.println("BREWING");
  tft.unloadFont();
}

void DisplayBrewComplete (void){ //UI Display Indicator that the Brew is Complete
  tft.fillScreen(TFT_GREEN);
  tft.setTextSize(1);
  tft.loadFont(Final_Frontier_28);
  tft.setCursor(30, 50, 4);

  tft.setTextColor(TFT_BLACK);
  tft.println("      BREW");
  tft.println("     COMPLETE");
  tft.unloadFont();
}

void DisplayBrewCompleteEnjoy(void){ // Adds enjoy to the above display function so it can blink on and off
  tft.fillScreen(TFT_GREEN);
  tft.setTextSize(1);
  tft.loadFont(Final_Frontier_28);
  tft.setCursor(30, 50, 4);

  tft.setTextColor(TFT_BLACK);
  tft.println("      BREW");
  tft.println("     COMPLETE");
  tft.unloadFont();
  tft.println();
  tft.setTextSize(2);
  tft.println("   ENJOY!");
  tft.setTextSize(1);
  tft.println("               Go Irish!");

}

void DisplayNeedMoreWater(void){  //UI Call when the Brew is aborted due to there being not enough water
  stopBrew();
  tft.fillScreen(TFT_CYAN);
  tft.setTextSize(1);
  tft.loadFont(Final_Frontier_28);
  tft.setCursor(30, 75, 4);
  
  tft.setTextColor(TFT_BLACK);
  tft.println(" ADD MORE");
  tft.println("     WATER TO");
  tft.println("    RESERVOIR");
  tft.println("         PLEASE");
  tft.unloadFont();
}

void DisplayCupRemoval(void){ // UI Call when the Brew is aborted due to the cup being removed
  tft.fillScreen(TFT_RED);
  tft.setTextSize(1);
  tft.loadFont(Final_Frontier_28);
  tft.setCursor(87, 70, 4);
  
  tft.setTextColor(TFT_WHITE);
  tft.println("CUP");
  tft.println("      REMOVED");
  tft.println("          BREW");
  tft.println("      ABORTED");
  tft.unloadFont();
}

void DisplayActiveFillRate(int dots, float liquidHeight, float targetHeight, float bottomHeight){ 
  //Actively Displays the Cup Fill % to indicate to user how much of their brew has been poured
  // int dots asks for input of how many dots to put after the word for the animation of the ...
  // Percentage is calculated (Current Pour Height - Inital Pour Height/Target Pour Height - Initial Pour Height)
  //Adjusts spacing for numbers on UI accordingly
  int iter;
  int dot_iter;
  float tempPercent = ((liquidHeight-bottomHeight)/(targetHeight-bottomHeight))*100.0;
  if(tempPercent > percent){
    percent = tempPercent;
  }
  //float LEDLight = (percent/100.0)*LEDHeightCounter;
  int percent_print = (int)percent;
  tft.setTextSize(1);
  tft.fillScreen(TFT_WHITE);
  tft.loadFont(Final_Frontier_28);
  tft.setCursor(0, 65, 4);
  tft.setTextColor(TFT_RED);
  tft.print("    POURING");
  if (dots != 0){
    for (dot_iter = 0; dot_iter < dots; dot_iter++){
      tft.print(" .");
    }
  }
  tft.println();
  tft.println();
  
  tft.setTextColor(TFT_BLUE);
  tft.unloadFont();
  tft.setTextSize(3);
  if (percent_print < 0){
    tft.print("    ");
    tft.print("0");    
  }
  else if (percent_print < 10){
    tft.print("    ");
    tft.print(percent_print);
  }
  else if (percent_print <= 100){
    tft.print("   ");
    tft.print(percent_print);
  }
  else{
    tft.print("  ");
    tft.print("100");
  }
  tft.print("%");
}

void DisplayHeatingInProgress(int dots){
  //Same as DisplayActiveFillRate except it changes color and text of TFT to indicate heating cycle, and doesn't display a fill %
  // int dots asks for input of how many dots to put after the word for the animation of the ...
  int iter;
  int dot_iter;
  tft.setTextSize(1);
  tft.fillScreen(TFT_ORANGE);
  tft.loadFont(Final_Frontier_28);
  tft.setCursor(0, 110, 4);
  tft.setTextColor(TFT_BLACK);
  tft.print("    HEATING");
  if (dots != 0){
    for (dot_iter = 0; dot_iter < dots; dot_iter++){
      tft.print(" .");
    }
  }
  tft.unloadFont();
}

void DisplayAddPod(void) //UI Display Call to indicate to the user that they should insert another K-Cup
{
  tft.fillScreen(TFT_YELLOW);
  tft.setTextSize(1);
  tft.loadFont(Final_Frontier_28);
  tft.setCursor(30, 50, 4);

  tft.setTextColor(TFT_BLACK);
  tft.println("   Large Brew");
  tft.println("         Detected");
  tft.println("    Please Replace");
  tft.println("  K-CUP and press");
  tft.println("    BREW button");
  tft.unloadFont();
}

void DisplayLogo(void){ // Displays Fill 'Er Up Logo
  tft.fillScreen(TFT_CYAN);
  tft.drawXBitmap(0, 0, FillErUpLogo, 240, 240, TFT_BLUE);
}

void DisplayNDEE(void){ // Displays ND EE Logo
  tft.fillScreen(TFT_GOLD);
  tft.drawXBitmap(0, 0, NDEE, 240, 240, TFT_NAVY);
}

void WaitStartUp(void){ //Loops through the Fill 'Er Up Logo and the ND EE Logo until a cup is sensed to be present
  //Displays the Fill 'Er Up Logo, continuously checks if the button is pressed for 2 seconds
  unsigned long timer = millis();
  while (1){
    timer = millis();
    DisplayLogo();
    while (millis() - timer < 3000){
      if (CheckCupPresent() == true){
        return;  //function can only be exited if the button is pressed
      }
    }
  //Displays the ND EE Logo, continuously checks if the button is pressed for 2 seconds
    timer = millis();
    DisplayNDEE();
    while (millis() - timer < 3000){
      if (CheckCupPresent() == true){
        return; //function can only be exited if the button is pressed
      }
    }
  }
}

void WaitRestart(void){
  //Loops through Brew Complete and Brew Complete, ENJOY! until either the user presses the dial to restart the process or the timer times out after a minute
  // and coffee maker goes into sleep mode
  unsigned long timer = millis();
  unsigned long timerstart = millis();
  while (millis() - timerstart <= 60000){
    DisplayBrewComplete();
    timer = millis();
    while (millis() - timer < 2000){
      if (ButtonCheck() == true){
        return;
      }
    }
    DisplayBrewCompleteEnjoy();
    timer = millis();
    while (millis() - timer < 2000){
      if (ButtonCheck() == true){
        return;
      }
    }
  }
}

void DisplayPowerOn(void){ //UI Display Call that prompts the user to turn the Keurig On
  tft.fillScreen(TFT_WHITE);
  tft.setTextSize(1);
  tft.loadFont(Final_Frontier_28);
  tft.setCursor(30, 90, 4);
  
  tft.setTextColor(TFT_BLACK);
  tft.println("     PLUG IN");
  tft.println("         KEURIG");
  tft.println("   TO PROCEED");
  tft.unloadFont();
}


void DisplayBrewAborted(void){ //UI Display Call that tells the user that the brew was aborted for some reason 
  tft.fillScreen(TFT_RED);
  tft.loadFont(Final_Frontier_28);
  tft.setCursor(0, 65, 4);

  tft.setTextColor(TFT_WHITE);
  tft.println("          BREW");
  tft.println("      ABORTED");
  tft.println("      Check Water");
  tft.println("       and Restart");
}

void DisplayMeasuring(void){ //UI Display Call indicating to the User that the Cup height TOF measure is currently occuring. 
  tft.fillScreen(TFT_DARKGREEN);
  tft.loadFont(Final_Frontier_28);
  tft.setCursor(0, 100, 4);
  tft.setTextColor(TFT_CYAN);
  tft.println("     MEASURING");
  tft.println("     CUP HEIGHT");
  tft.unloadFont();
}





